import os
import pathlib

import pandas as pd
from plotnine import aes, geom_line, ggplot, scale_color_manual

from . import utils

grid_colors = [
    "#FF0000",
    "#00FF00",
    "#0000FF",
    "#FFFF00",
    "#FF00FF",
    "#00FFFF",
    "#000000",
    "#888888",
    "#888800",
    "#008888",
    "#880088",
    "#8888FF",
    "#FF8888",
    "#88FF88",
    "#880000",
    "#88FF00",
    "#88FFFF",
    "#8800FF",
    "#008800",
    "#FF8800",
    "#FF88FF",
    "#0088FF",
    "#000088",
    "#FF0088",
    "#FFFF88",
    "#00FF88",
]


def mean_over_up_del_size_on_tem(root_dir: os.PathLike, target: str, stem_wise: bool):
    root_dir = pathlib.Path(os.fspath(root_dir))
    df_treat = (
        pd.read_feather(root_dir / "analyze" / "treat" / "correct" / "treat.feather")
        .query("tem_indicator")
        .reset_index(drop=True)
    )
    for tem in range(1, 5):
        save_dir = (
            root_dir
            / "figures"
            / "analyze"
            / "mean_freq_over_up_del_size_on_tem"
            / target
            / str(tem)
        )
        os.makedirs(save_dir, exist_ok=True)
        mean_over_up_del_size_on_tem_inner(
            df_treat,
            tem,
            target,
            save_dir / "all.pdf",
        )
        if stem_wise:
            for stem in df_treat["stem"].unique().tolist():
                df_treat_stem = df_treat.query("stem == @stem").reset_index(drop=True)
                mean_over_up_del_size_on_tem_inner(
                    df_treat_stem,
                    tem,
                    target,
                    save_dir / f"{stem}.pdf",
                )


def mean_over_up_del_size_on_tem_inner(
    df_treat: pd.DataFrame, tem: int, target: str, save_file: os.PathLike
):
    df_tem = (
        df_treat.query(
            """
                cut2 - ref_start2 == @tem and \
                ref_end1 <= cut1
            """
        )
        .reset_index(drop=True)
        .assign(
            up_del_size=lambda df: df["cut1"] - df["ref_end1"],
        )
    )

    if not df_tem["legal"].any():
        return

    df_value = (
        utils.pivot_value(df=df_tem, value=target, column="up_del_size")
        .assign(cas=lambda df: df["stem"].map(utils.infer_cas))
        .drop(columns=["stem", "ref_id"])
        .groupby("cas")
        .sum()
    )
    df_legal = (
        utils.pivot_legal(df=df_tem, column="up_del_size")
        .assign(cas=lambda df: df["stem"].map(utils.infer_cas))
        .drop(columns=["stem", "ref_id"])
        .groupby("cas")
        .sum()
    )
    value_vars = df_legal.columns

    df_mean = (
        (df_value / df_legal)
        .reset_index()
        .melt(
            id_vars="cas",
            value_vars=value_vars,
            var_name="up_del_size",
            value_name=target,
        )
    )

    (
        ggplot(df_mean, aes(x="up_del_size", y=target, group="cas", color="cas"))
        + geom_line()
        + scale_color_manual(values=grid_colors)
    ).save(save_file)



import os
import pathlib
import re
import subprocess
import tempfile
from typing import Iterable

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from Bio import Seq


def infer_cas(file: os.PathLike) -> str:
    name = pathlib.Path(os.fspath(file)).name.lower()
    if (
        name.startswith("a2")
        or name.startswith("a7")
        or name.startswith("d2")
        or name.startswith("h73")
        or name.startswith("h80")
        or name.startswith("p79")
        or name.startswith("p125")
        or name.startswith("q101")
        or name.startswith("q102")
        or name.startswith("q192")
    ):
        return "spycas9"
    if name.startswith("36t") or name.startswith("b2") or name.startswith("x"):
        return "spymac"
    if name.startswith("i10t") or name.startswith("i83"):
        return "ispymac"
    if name.startswith("wt"):
        return "control"
    raise Exception("unknown cas")


def infer_polq(file: os.PathLike) -> str:
    name = pathlib.Path(os.fspath(file)).name.lower()
    if (
        name.startswith("a2")
        or name.startswith("a7")
        or name.startswith("d2")
        or name.startswith("36t")
        or name.startswith("b2")
        or name.startswith("x")
        or name.startswith("i10t")
        or name.startswith("i83")
        or name.startswith("wt")
    ):
        return "wt"
    if name.startswith("h73") or name.startswith("h80"):
        return "hel"
    if name.startswith("p79") or name.startswith("p125"):
        return "pol"
    if name.startswith("q101") or name.startswith("q102") or name.startswith("q192"):
        return "ko"
    raise Exception("unknown polq")


def infer_chip(file: os.PathLike) -> str:
    name = pathlib.Path(os.fspath(file)).name.lower()
    return re.search(r"-(a1|a2|a3|g1n|g2n|g3n)-", name).group(1)


def infer_time(file: os.PathLike) -> int:
    name = pathlib.Path(os.fspath(file)).name.lower()
    return int(re.search(r"-(?:a1|a2|a3|g1n|g2n|g3n)-(\d)", name).group(1))


def infer_sample(file: os.PathLike) -> str:
    name = pathlib.Path(os.fspath(file)).name.lower()
    return re.search(r"^(.+?)-", name).group(1)


def infer_rep(file: os.PathLike) -> int:
    name = pathlib.Path(os.fspath(file)).name.lower()
    if re.search(r"re21-", name):
        return 3
    if (
        name.startswith("d21-")
        or name.startswith("a71-")
        or name.startswith("wt11-")
        or name.startswith("wt21-")
        or re.search(r"re2-", name)
    ):
        return 2
    return 1


def infer_wt(file: os.PathLike) -> str:
    assert infer_cas(file) == "control", "not wild type"
    name = pathlib.Path(os.fspath(file)).name.lower()
    wt = name.split("-", 1)[0]
    return wt


def p5primer() -> str:
    return "GTGGAAAGGACGAAACACC"


def p7primer() -> str:
    return str(Seq.Seq("AATTCGCTAGCTAGGTCTTGA").reverse_complement())


def normal_scaffold() -> str:
    return "GTTTTAGAGCTAGAAATAGCAAGTTAAAATAAGGCTAGTCCGTTATCAACTTGAAAAAGTGGCACCGAGTCGGTGCTTTTTT"


def nbt_scaffold() -> str:
    return "GTTTCAGAGCTATGCTGGAAACAGCATAGCAAGTTGAAATAAGGCTAGTCCGTTATCAACTTGAAAAAGTGGCACCGAGTCGGTGCTTTTTT"


def scaffold(chip: str) -> str:
    if chip in ["a1", "a2", "a3"]:
        return nbt_scaffold()
    else:
        return normal_scaffold()


def scaffold_alt(chip: str) -> str:
    if chip in ["a1", "a2", "a3"]:
        return normal_scaffold()
    else:
        return nbt_scaffold()


def RCscaffold(chip: str) -> str:
    return str(Seq.Seq(scaffold(chip)).reverse_complement())


def RCscaffold_alt(chip: str) -> str:
    return str(Seq.Seq(scaffold_alt(chip)).reverse_complement())


def rev_comp(seq: str) -> str:
    return str(Seq.Seq(seq).reverse_complement())


#################################################
# analyze
#################################################
def up_del_size(df: pd.DataFrame) -> pd.Series:
    up_del_size: pd.Series = np.maximum(df["cut1"] - df["ref_end1"], 0)
    return up_del_size


def down_del_size(df: pd.DataFrame) -> pd.Series:
    down_del_size: pd.Series = np.maximum(df["ref_start2"] - df["cut2"], 0)
    return down_del_size


def rand_ins_size(df: pd.DataFrame) -> pd.Series:
    rand_ins_size: pd.Series = df["random_insertion"].str.len()
    return rand_ins_size


def is_wt(df: pd.DataFrame) -> pd.Series:
    return (
        (df["ref_end1"] == df["cut1"])
        & (df["ref_start2"] == df["cut2"])
        & (df["random_insertion"] == "")
    )


def freq_mutant(df: pd.DataFrame) -> pd.Series:
    freq_mutant = df["count"] / df.groupby(["stem", "ref_id"])["count"].transform("sum")
    freq_mutant.loc[is_wt(df)] = float("nan")

    return freq_mutant


def count_wt(df: pd.DataFrame) -> pd.Series:
    return (
        df[["stem", "ref_id"]]
        .merge(
            right=df.loc[is_wt(df)][["stem", "ref_id", "count"]],
            how="left",
            on=["stem", "ref_id"],
            validate="many_to_one",
        )["count"]
        .fillna(0)
    )


def freq_nowt(df: pd.DataFrame) -> pd.Series:
    freq_nowt = 1 - count_wt(df) / df.groupby(["stem", "ref_id"])["count"].transform(
        "sum"
    )

    return freq_nowt


def count_tem_dele(df: pd.DataFrame, tem: int, dele: int) -> pd.Series:
    return (
        df[["stem", "ref_id", "random_insertion"]]
        .merge(
            right=df.query(
                """
                    ref_end1 + @dele == cut1 and \
                    ref_start2 + @tem == cut2
                """
            )[["stem", "ref_id", "random_insertion", "count"]],
            how="left",
            on=["stem", "ref_id", "random_insertion"],
            validate="many_to_one",
        )["count"]
        .fillna(0)
    )


def kim(df: pd.DataFrame) -> tuple[pd.Series]:
    count_tot = df.groupby(["stem", "ref_id"])["count"].transform("sum")
    count_kim = df["count"] - count_tot * (
        df["count_ctl"] / (df["count_tot_ctl"] + 1e-6)
    )
    count_kim[is_wt(df)] = float("nan")
    count_kim_sum = (
        df.assign(count_kim=count_kim)
        .groupby(["stem", "ref_id"])["count_kim"]
        .transform("sum")
    )
    # Add a small value to both count_tot_ctl and count_wt_ctl so that if count_tot_ctl is 0, then no correction will be applied.
    count_tot_kim = count_tot * (
        (df["count_wt_ctl"] + 1e-6) / (df["count_tot_ctl"] + 1e-6)
    )
    freq_mut_kim = count_kim_sum / (count_tot_kim + 1e-6)

    return count_kim, count_tot_kim, freq_mut_kim


def pivot_value(df: pd.DataFrame, value: str, column: str) -> pd.DataFrame:
    df = (
        df.assign(**{value: lambda df: df[value].where(df["legal"], float("nan"))})
        .pivot_table(
            values=value,
            index=["stem", "ref_id"],
            columns=column,
            aggfunc="sum",
            fill_value=0,
        )
        .reset_index()
    )

    return df


def pivot_legal(df: pd.DataFrame, column: str) -> pd.DataFrame:
    df = df.pivot_table(
        values="legal",
        index=["stem", "ref_id"],
        columns=column,
        aggfunc="any",
        fill_value=True,
    ).reset_index()

    return df


###########################################
# rearr
###########################################


def read_alg(alg_file: os.PathLike, names: list[str]) -> pd.DataFrame:
    df_alg = pd.read_csv(alg_file, header=None).loc[:, 0]
    df_alg.index = pd.MultiIndex.from_tuples(
        zip(
            np.repeat(np.arange(df_alg.shape[0] // 3), 3),
            np.tile(["meta", "ref", "query"], df_alg.shape[0] // 3),
        )
    )
    df_alg = df_alg.unstack()
    df_meta = (
        df_alg["meta"]
        .str.split("\t", expand=True)
        .rename(columns={idx: name for idx, name in enumerate(names)})
    ).astype(
        {
            name: int if name != "count_distri" else float
            for name in names
            if name not in ["updangle", "random_insertion", "downdangle"]
        }
    )
    df_alg = pd.concat([df_meta, df_alg.drop(columns="meta")], axis=1)

    return df_alg


def call_rearr(
    ref1s: Iterable[str],
    ref2s: Iterable[str],
    cuts: Iterable[int],
    exts: Iterable[int],
    queries: Iterable[str],
    counts: Iterable[int],
) -> pd.DataFrame:
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = pathlib.Path(os.fspath(tmpdir))
        with (
            open(tmpdir / "ref", "w") as fd_ref,
            open(tmpdir / "query", "w") as fd_query,
            open(tmpdir / "correct", "w") as fd_correct,
        ):
            for ref_id, (ref1, ref2, cut, ext, query, count) in enumerate(
                zip(ref1s, ref2s, cuts, exts, queries, counts)
            ):
                fd_ref.write(f"{0}\t{ref1}\t{cut}\t{ext}\t{ref2}\t{len(ref2)}\n")
                fd_query.write(f"{query}\t{count}\t{ref_id}\n")
                fd_correct.write(f"up\n")

        subprocess.run(
            args=" ".join(
                [
                    "rearrangement",
                    "<",
                    (tmpdir / "query").as_posix(),
                    "3<",
                    (tmpdir / "ref").as_posix(),
                    "|",
                    "correct_micro_homology.awk",
                    "--",
                    (tmpdir / "ref").as_posix(),
                    (tmpdir / "correct").as_posix(),
                    ">",
                    (tmpdir / "alg").as_posix(),
                ]
            ),
            shell=True,
        )

        df_alg = read_alg(
            tmpdir / "alg",
            names=[
                "index",
                "count",
                "score",
                "ref_id",
                "updangle",
                "ref_start1",
                "query_start1",
                "ref_end1",
                "query_end1",
                "random_insertion",
                "ref_start2",
                "query_start2",
                "ref_end2",
                "query_end2",
                "downdangle",
                "cut1",
                "cut2",
            ],
        )

    return df_alg


def infer_cut(row: pd.Series, ext: int) -> pd.Series:
    ref_line = row["ref"]
    query_line = row["query"]
    ref_end1 = row["ref_end1"]
    ref_start2 = row["ref_start2"]
    cut = row["cut_query"]
    first_lower_end = re.search(r"[acgtn]", ref_line).span()[1]
    before_rand_ins = (
        re.search(r"[acgtn]", ref_line[first_lower_end:]).span()[1] + first_lower_end
    )
    after_rand_ins = (
        re.search(r"[acgtn]", ref_line[before_rand_ins:]).span()[0] + before_rand_ins
    )
    ref_line_array = np.array(list(ref_line))
    temp = (ref_line_array[:before_rand_ins] != "-").cumsum()
    ref_end1_alg = np.where(temp == ref_end1)[0].item() + 1
    ref1_len = temp[-1]
    ref_start2_alg = (
        np.where(
            (ref_line_array[after_rand_ins:] != "-").cumsum() == ref_start2 - ref1_len
        )[0].item()
        + 1
        + after_rand_ins
    )

    query_line_array = np.array(list(query_line))
    cut_alg = np.where((query_line_array != "-").cumsum() == cut)[0].item() + 1
    if cut_alg <= ref_end1_alg:
        ref1_correct = ref_line[:cut_alg]
    elif cut_alg <= after_rand_ins:
        ref1_correct = ref_line[:ref_end1_alg] + query_line[before_rand_ins:cut_alg]
    else:
        ref1_correct = (
            ref_line[:ref_end1_alg]
            + query_line[before_rand_ins:after_rand_ins]
            + ref_line[ref_start2_alg:cut_alg]
        )

    cut_correct = len(ref1_correct.replace("-", ""))
    ref_correct = (
        (
            ref_line[:ref_end1_alg]
            + query_line[before_rand_ins:after_rand_ins]
            + ref_line[ref_start2_alg:]
        )
        .replace("-", "")
        .upper()
    )
    assert cut_correct >= ext, "ref1 too short"
    assert len(ref_correct) - cut_correct >= ext, "ref2 too short"
    ref1_correct = ref_correct[: cut_correct + ext]
    ref2_correct = ref_correct[cut_correct - ext :]

    return pd.Series(
        {
            "ref1": ref1_correct,
            "ref2": ref2_correct,
            "cut": cut_correct,
        }
    )


def get_mh(row: pd.Series) -> str:
    ref_end1 = row["ref_end1"]
    cut1 = row["cut1"]
    ref1 = row["ref1"]
    ref_start2 = row["ref_start2"] - len(ref1)
    cut2 = row["cut2"] - len(ref1)
    ref2 = row["ref2"]
    i = ref_end1
    for i in range(ref_end1, cut1 + 1):
        if i >= cut1 or ref_start2 + i - ref_end1 >= len(ref2):
            break
        if ref1[i] != ref2[ref_start2 + i - ref_end1]:
            break
    j = ref_start2 - 1
    for j in range(ref_start2 - 1, cut2 - 2, -1):
        if j < cut2 or ref_end1 - ref_start2 + j < 0:
            break
        if ref2[j] != ref1[ref_end1 - ref_start2 + j]:
            break

    return ref2[j + 1 : ref_start2] + ref1[ref_end1:i]



import os
import pathlib

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import patsy
import statsmodels.api as sm
from scipy import special

from . import utils


def correct_alg(root_dir: os.PathLike, correct_dir: os.PathLike, temperature: float):
    root_dir = pathlib.Path(os.fspath(root_dir))
    correct_dir = pathlib.Path(os.fspath(correct_dir))
    os.makedirs(correct_dir, exist_ok=True)
    for alg_file in os.listdir(root_dir / "align" / "raw"):
        df_alg = utils.read_alg(
            root_dir / "align" / "raw" / alg_file,
            names=[
                "index",
                "count",
                "score",
                "ref_id",
                "updangle",
                "ref_start1",
                "query_start1",
                "ref_end1",
                "query_end1",
                "random_insertion",
                "ref_start2",
                "query_start2",
                "ref_end2",
                "query_end2",
                "downdangle",
                "cut1",
                "cut2",
            ],
        )
        df_query = pd.read_csv(
            root_dir / "query" / "found" / f"{pathlib.Path(alg_file).stem}.query",
            sep="\t",
            names=["query", "count", "ref_id"],
            keep_default_na=False,
        )
        df_alg["index"] = df_query.groupby("query").transform("ngroup")

        chip = utils.infer_chip(alg_file)
        df_ref = pd.read_feather(
            root_dir / "control" / "hq_mut" / f"{chip}.feather"
        ).reset_index(names="ref_id")
        df_alg = df_alg.merge(
            right=df_ref[["ref_id", "count"]].rename(columns={"count": "count_ref"}),
            how="left",
            on=["ref_id"],
            validate="many_to_one",
        ).assign(
            weight=lambda df: (
                df["count_ref"]
                * df.groupby("index")["score"].transform(
                    lambda score, temperature=temperature: special.softmax(
                        score / temperature
                    )
                )
            ),
            count_distri=lambda df: df["count"]
            * df.groupby("index")["weight"].transform(lambda se: se / se.sum()),
        )

        df_alg = df_alg.astype(str)
        df_output = pd.DataFrame(
            {
                "meta": df_alg["index"].str.cat(
                    df_alg[
                        [
                            "count",
                            "score",
                            "ref_id",
                            "updangle",
                            "ref_start1",
                            "query_start1",
                            "ref_end1",
                            "query_end1",
                            "random_insertion",
                            "ref_start2",
                            "query_start2",
                            "ref_end2",
                            "query_end2",
                            "downdangle",
                            "cut1",
                            "cut2",
                            "count_ref",
                            "count_distri",
                        ]
                    ],
                    sep="\t",
                ),
                "ref": df_alg["ref"],
                "query": df_alg["query"],
            },
        )
        df_output.stack().to_csv(
            correct_dir / alg_file,
            header=False,
            index=False,
        )


def stat_read(root_dir: os.PathLike, correct_dir: os.PathLike):
    root_dir = pathlib.Path(os.fspath(root_dir))
    correct_dir = pathlib.Path(os.fspath(correct_dir))
    save_dir = root_dir / "figures" / "analyze" / "stat_read"
    os.makedirs(save_dir, exist_ok=True)
    df_algs = []
    for alg_file in os.listdir(correct_dir):
        df_algs.append(
            utils.read_alg(
                correct_dir / alg_file,
                names=[
                    "index",
                    "count",
                    "score",
                    "ref_id",
                    "updangle",
                    "ref_start1",
                    "query_start1",
                    "ref_end1",
                    "query_end1",
                    "random_insertion",
                    "ref_start2",
                    "query_start2",
                    "ref_end2",
                    "query_end2",
                    "downdangle",
                    "cut1",
                    "cut2",
                    "count_ref",
                    "count_distri",
                ],
            )
            .groupby("score")["count_distri"]
            .sum()
            .rename("count")
            .reset_index()
        )

    df_alg = pd.concat(df_algs)
    df_alg["score"].plot.hist(bins=300, weights=df_alg["count"]).get_figure().savefig(
        save_dir / "score.pdf"
    )
    plt.close("all")


def collect_data(
    root_dir: os.PathLike,
    correct_dir: os.PathLike,
    min_score: int,
):
    """
    Only collect data. Do not apply any annotation. Only apply read-wise filter such as score.
    """
    root_dir = pathlib.Path(os.fspath(root_dir))
    correct_dir = pathlib.Path(os.fspath(correct_dir))
    os.makedirs(root_dir / "analyze" / "treat" / "full", exist_ok=True)
    os.makedirs(root_dir / "analyze" / "control" / "full", exist_ok=True)
    os.makedirs(root_dir / "figures" / "analyze" / "collect_data", exist_ok=True)

    df_stat = pd.DataFrame(
        {
            "row_num": [0, 0],
            "count": [0, 0],
        },
        index=["full", "filter"],
    )
    df_algs = []
    for alg_file in os.listdir(correct_dir):
        df_alg = utils.read_alg(
            correct_dir / alg_file,
            names=[
                "index",
                "count",
                "score",
                "ref_id",
                "updangle",
                "ref_start1",
                "query_start1",
                "ref_end1",
                "query_end1",
                "random_insertion",
                "ref_start2",
                "query_start2",
                "ref_end2",
                "query_end2",
                "downdangle",
                "cut1",
                "cut2",
                "count_ref",
                "count_distri",
            ],
        )

        df_stat.loc["full", "row_num"] += df_alg.shape[0]
        df_stat.loc["full", "count"] += df_alg["count_distri"].sum()

        df_alg = df_alg.query("score >= @min_score")

        df_stat.loc["filter", "row_num"] += df_alg.shape[0]
        df_stat.loc["filter", "count"] += df_alg["count_distri"].sum()

        df_alg = (
            df_alg.groupby(
                [
                    "ref_id",
                    "cut1",
                    "cut2",
                    "ref_end1",
                    "random_insertion",
                    "ref_start2",
                ]
            )["count_distri"]
            .sum()
            .rename("count")
            .reset_index()
            .astype(
                {
                    "ref_end1": "int8",
                    "ref_start2": "int8",
                    "cut1": "int8",
                    "cut2": "int8",
                }
            )
            .assign(stem=pathlib.Path(alg_file).stem)
        )
        df_algs.append(df_alg)

    df_stat.to_csv(root_dir / "figures" / "analyze" / "collect_data" / "stat.csv")
    df_stat["row_num"].plot.bar().get_figure().savefig(
        root_dir / "figures" / "analyze" / "collect_data" / "row_num.pdf"
    )
    plt.close("all")
    df_stat["count"].plot.bar().get_figure().savefig(
        root_dir / "figures" / "analyze" / "collect_data" / "count.pdf"
    )
    plt.close("all")

    df_alg = pd.concat(df_algs).assign(cas=lambda df: df["stem"].map(utils.infer_cas))
    df_alg.query("cas != 'control'").drop(columns="cas").reset_index(
        drop=True
    ).to_feather(root_dir / "analyze" / "treat" / "full" / "treat.feather")
    df_alg.query("cas == 'control'").drop(columns="cas").reset_index(
        drop=True
    ).to_feather(root_dir / "analyze" / "control" / "full" / "control.feather")


def stat_ref(root_dir: os.PathLike, min_count_tot: int, max_up_del_size: int):
    root_dir = pathlib.Path(os.fspath(root_dir))
    save_dir = root_dir / "figures" / "analyze" / "stat_ref"
    os.makedirs(save_dir, exist_ok=True)

    df_treat = pd.read_feather(
        root_dir / "analyze" / "treat" / "full" / "treat.feather"
    )
    df_treat.groupby(["stem", "ref_id"])["count"].sum().clip(upper=300).plot.hist(
        bins=np.linspace(0, 301, 302)
    ).get_figure().savefig(save_dir / "count_tot.pdf")
    plt.close("all")

    df_treat = df_treat.assign(
        count_tot=lambda df: df.groupby(["stem", "ref_id"])["count"].transform("sum")
    )

    df_treat.assign(freq_nowt=utils.freq_nowt(df_treat)).query(
        "count_tot >= @min_count_tot"
    ).groupby(["stem", "ref_id"])["freq_nowt"].first().plot.hist(
        bins=100
    ).get_figure().savefig(
        save_dir / "freq_nowt.pdf"
    )
    plt.close("all")

    for tem in range(1, 5):
        df_treat_tem = df_treat.assign(
            count_blunt=lambda df: utils.count_tem_dele(df, tem, 0),
            freq_blunt=lambda df: df["count_blunt"] / (df["count_tot"] + 1e-6),
        )
        for dele in range(1, max_up_del_size + 1):
            df_treat_dele = (
                df_treat_tem.assign(
                    count_dele=lambda df: utils.count_tem_dele(df, tem, dele),
                    freq_dele=lambda df: df["count_dele"] / (df["count_tot"] + 1e-6),
                    freq_dele_rel_blunt=lambda df: df["count_dele"]
                    / (df["count_blunt"] + 1e-6),
                )
                .query("count_tot >= @min_count_tot")
                .groupby(["stem", "ref_id"])
                .agg(
                    freq_blunt=pd.NamedAgg(column="freq_blunt", aggfunc="first"),
                    freq_dele=pd.NamedAgg(column="freq_dele", aggfunc="first"),
                    freq_dele_rel_blunt=pd.NamedAgg(
                        column="freq_dele_rel_blunt", aggfunc="first"
                    ),
                )
            )

            for column in ["freq_blunt", "freq_dele", "freq_dele_rel_blunt"]:
                if column == "freq_blunt":
                    name = f"freq_blunt_tem{tem}"
                elif column == "freq_dele":
                    name = f"freq_dele{dele}_tem{tem}"
                elif column == "freq_dele_rel_blunt":
                    name = f"freq_dele{dele}_tem{tem}_rel_blunt"
                df_treat_dele[column].clip(upper=10).plot.hist(
                    bins=300, logy=True
                ).get_figure().savefig(save_dir / f"{name}.pdf")
                plt.close("all")


def fit_ref(root_dir: os.PathLike, max_up_del_size: int):
    root_dir = pathlib.Path(os.fspath(root_dir))
    os.makedirs(root_dir / "fit", exist_ok=True)
    save_dir = root_dir / "figures" / "analyze" / "fit_ref"
    os.makedirs(save_dir, exist_ok=True)

    df_treat = pd.read_feather(
        root_dir / "analyze" / "treat" / "full" / "treat.feather"
    )
    df_treat = (
        df_treat.assign(
            up_del_size=lambda df: utils.up_del_size(df),
        )
        .query("ref_end1 <= cut1 and up_del_size <= @max_up_del_size")
        .groupby(["up_del_size"])["count"]
        .sum()
        .reset_index()
        .assign(
            freq_rel_blunt=lambda df: df["count"]
            / df.loc[df["up_del_size"] == 0, "count"].item()
        )
    )

    log_freq_rel_blunt, const_up_del_size = patsy.dmatrices(
        "np.log(freq_rel_blunt) ~ up_del_size", data=df_treat, return_type="dataframe"
    )
    result = sm.OLS(
        endog=log_freq_rel_blunt,
        exog=const_up_del_size,
    ).fit()
    prediction = result.get_prediction(const_up_del_size)
    summary_frame = prediction.summary_frame(alpha=0.05)
    summary_frame = summary_frame.assign(
        up_del_size=const_up_del_size["up_del_size"],
    )
    summary_frame.to_csv(root_dir / "fit" / "summary_frame.csv", index=False)

    plt.scatter(x=df_treat["up_del_size"], y=np.log(df_treat["freq_rel_blunt"]))
    plt.plot(
        summary_frame["up_del_size"],
        summary_frame["mean"],
        color="red",
        label="fitted freq_rel_blunt",
    )
    plt.fill_between(
        summary_frame["up_del_size"],
        summary_frame["mean_ci_lower"],
        summary_frame["mean_ci_upper"],
        color="red",
        alpha=0.3,
        label="95% CI",
    )
    plt.fill_between(
        summary_frame["up_del_size"],
        summary_frame["obs_ci_lower"],
        summary_frame["obs_ci_upper"],
        color="blue",
        alpha=0.2,
        label="95% PI",
    )
    plt.savefig(save_dir / "ci.pdf")
    plt.close("all")


def filter_ref(
    root_dir: os.PathLike,
    min_count_tot: int,
    max_freq_nowt: float,
):
    """
    Use positive mask because nan compare always return False.
    """
    root_dir = pathlib.Path(os.fspath(root_dir))
    os.makedirs(root_dir / "analyze" / "treat" / "filter" / "ref", exist_ok=True)
    save_dir = root_dir / "figures" / "analyze" / "filter_ref"
    os.makedirs(save_dir, exist_ok=True)

    df_treat = pd.read_feather(
        root_dir / "analyze" / "treat" / "full" / "treat.feather"
    )

    df_stat = pd.DataFrame(index=["full", "filter"], columns=["ref_num", "count"])
    df_stat.loc["full", "ref_num"] = (
        df_treat[["stem", "ref_id"]].drop_duplicates().shape[0]
    )
    df_stat.loc["full", "count"] = df_treat["count"].sum()

    count_tot = df_treat.groupby(["stem", "ref_id"])["count"].transform("sum")
    mask = count_tot >= min_count_tot
    mask = mask & (utils.freq_nowt(df_treat) <= max_freq_nowt)

    df_treat = df_treat.loc[mask].reset_index(drop=True)

    df_stat.loc["filter", "ref_num"] = (
        df_treat[["stem", "ref_id"]].drop_duplicates().shape[0]
    )
    df_stat.loc["filter", "count"] = df_treat["count"].sum()
    df_stat.to_csv(save_dir / "stat.csv")
    df_stat["ref_num"].plot.bar().get_figure().savefig(save_dir / "ref_num.pdf")
    plt.close("all")
    df_stat["count"].plot.bar().get_figure().savefig(save_dir / "count.pdf")
    plt.close("all")

    df_treat.to_feather(
        root_dir / "analyze" / "treat" / "filter" / "ref" / "treat.feather"
    )


def stat_mutant(root_dir: os.PathLike, min_count_tot: int):
    root_dir = pathlib.Path(os.fspath(root_dir))
    save_dir = root_dir / "figures" / "analyze" / "stat_mutant"
    os.makedirs(save_dir, exist_ok=True)

    df_treat = pd.read_feather(
        root_dir / "analyze" / "treat" / "filter" / "ref" / "treat.feather"
    )

    utils.up_del_size(df_treat).clip(upper=30).plot.hist(
        bins=np.linspace(0, 31, 32), weights=df_treat["count"]
    ).get_figure().savefig(save_dir / "up_del_size.pdf")
    plt.close("all")

    utils.down_del_size(df_treat).clip(upper=30).plot.hist(
        bins=np.linspace(0, 31, 32), weights=df_treat["count"]
    ).get_figure().savefig(save_dir / "down_del_size.pdf")
    plt.close("all")

    utils.rand_ins_size(df_treat).clip(upper=30).plot.hist(
        bins=np.linspace(0, 31, 32), weights=df_treat["count"]
    ).get_figure().savefig(save_dir / "rand_ins_size.pdf")
    plt.close("all")

    df_treat.assign(
        count_tot=lambda df: df.groupby(["stem", "ref_id"])["count"].transform("sum"),
        freq_mutant=lambda df: utils.freq_mutant(df),
    ).query("count_tot >= @min_count_tot")["freq_mutant"].plot.hist(
        bins=100
    ).get_figure().savefig(
        save_dir / "freq_mutant.pdf"
    )
    plt.close("all")


def filter_mutant(
    root_dir: os.PathLike,
    max_up_del_size: int,
    max_down_del_size: int,
    max_rand_ins_size: int,
    max_freq_mutant: float,
    fit_tem_max: int,
):
    """
    Do not filter mutant because missing mutant are treated as count 0. Use a column legal to mark mutant passing the filter.
    """
    root_dir = pathlib.Path(os.fspath(root_dir))
    os.makedirs(root_dir / "analyze" / "treat" / "filter" / "mutant", exist_ok=True)
    save_dir = root_dir / "figures" / "analyze" / "filter_mutant"
    os.makedirs(save_dir, exist_ok=True)

    df_treat = pd.read_feather(
        root_dir / "analyze" / "treat" / "filter" / "ref" / "treat.feather"
    )

    df_stat = pd.DataFrame(index=["full", "filter"], columns=["mutant_num", "count"])
    df_stat.loc["full", "mutant_num"] = df_treat.shape[0]
    df_stat.loc["full", "count"] = df_treat["count"].sum()

    outlier_ratio = np.exp(
        pd.read_csv(root_dir / "fit" / "summary_frame.csv", header=0)["obs_ci_upper"]
    )
    df_treat["tem_indicator"] = df_treat["cut2"] - df_treat["ref_start2"] > fit_tem_max
    for tem in range(1, fit_tem_max + 1):
        df_treat["tem_indicator"] = df_treat["tem_indicator"] | (
            (df_treat["cut2"] - df_treat["ref_start2"] == tem)
            & (df_treat["cut1"] - df_treat["ref_end1"] == 0)
        )
        count_blunt = utils.count_tem_dele(df_treat, tem, 0)
        for up_del_size in range(1, max_up_del_size + 1):
            df_treat["tem_indicator"] = df_treat["tem_indicator"] | (
                (df_treat["cut2"] - df_treat["ref_start2"] == tem)
                & (df_treat["cut1"] - df_treat["ref_end1"] == up_del_size)
                & (
                    utils.count_tem_dele(df_treat, tem, up_del_size)
                    <= outlier_ratio[up_del_size] * count_blunt
                )
            )

    df_treat = df_treat.assign(
        legal=(
            (utils.up_del_size(df_treat) <= max_up_del_size)
            & (utils.down_del_size(df_treat) <= max_down_del_size)
            & (utils.rand_ins_size(df_treat) <= max_rand_ins_size)
            & ((utils.freq_mutant(df_treat) <= max_freq_mutant) | utils.is_wt(df_treat))
        )
    )

    df_stat.loc["filter", "mutant_num"] = df_treat["legal"].sum()
    df_stat.loc["filter", "count"] = df_treat.query("legal")["count"].sum()
    df_stat.to_csv(save_dir / "stat.csv")
    df_stat["mutant_num"].plot.bar().get_figure().savefig(save_dir / "mutant_num.pdf")
    plt.close("all")
    df_stat["count"].plot.bar().get_figure().savefig(save_dir / "count.pdf")
    plt.close("all")

    df_treat.to_feather(
        root_dir / "analyze" / "treat" / "filter" / "mutant" / "treat.feather"
    )


def kim_correct(root_dir: os.PathLike):
    root_dir = pathlib.Path(os.fspath(root_dir))
    os.makedirs(root_dir / "analyze" / "treat" / "correct", exist_ok=True)

    df_treat = pd.read_feather(
        root_dir / "analyze" / "treat" / "merge" / "treat.feather"
    )
    count_kim, count_tot_kim, freq_mut_kim = utils.kim(df_treat)
    df_treat = df_treat.assign(
        count_kim=count_kim,
        count_tot_kim=count_tot_kim,
        freq_mut_kim=freq_mut_kim,
    )

    df_treat.to_feather(
        root_dir / "analyze" / "treat" / "correct" / "treat.feather",
    )


def annote_columns(root_dir: os.PathLike):
    root_dir = pathlib.Path(os.fspath(root_dir))
    os.makedirs(root_dir / "analyze" / "treat" / "annote", exist_ok=True)
    os.makedirs(root_dir / "analyze" / "control" / "annote", exist_ok=True)

    df_treat = pd.read_feather(
        root_dir / "analyze" / "treat" / "correct" / "treat.feather"
    )
    df_control = pd.read_feather(
        root_dir / "analyze" / "control" / "dup" / "control.feather"
    )

    # Read barcode and sgRNA from the generated fasta file to prevent the inconsistency when use custom plasmid file.
    df_bar = (
        pd.read_csv(root_dir / "barcode" / "index" / "barcode.fa", names=["barcode"])
        .loc[1::2, :]
        .reset_index(drop=True)
        .reset_index(names="barcode_id")
    )
    df_sgRNA = (
        pd.read_csv(root_dir / "sgRNA" / "index" / "sgRNA.fa", names=["sgRNA"])
        .loc[1::2, :]
        .reset_index(drop=True)
        .reset_index(names="barcode_id")
    )
    df_ref = []
    for chip in ["a1", "a2", "a3", "g1n", "g2n", "g3n"]:
        df_ref.append(
            pd.read_csv(
                root_dir / "ref" / f"{chip}.ref",
                sep="\t",
                names=["zero", "ref1", "cut1", "ext", "ref2", "ref2len"],
            )[["ref1", "ref2"]]
            .reset_index(names="ref_id")
            .assign(chip=chip)
        )
    df_ref = pd.concat(df_ref).reset_index(drop=True)

    df_column = []
    for chip in ["a1", "a2", "a3", "g1n", "g2n", "g3n"]:
        df_column.append(
            pd.read_feather(root_dir / "control" / "hq_mut" / f"{chip}.feather")[
                ["barcode_id"]
            ]
            .reset_index(names="ref_id")
            .assign(chip=chip)
        )
    df_column = pd.concat(df_column).reset_index(drop=True)

    df_column = (
        df_column.merge(
            right=df_bar,
            how="left",
            on=["barcode_id"],
            validate="many_to_one",
        )
        .merge(
            right=df_sgRNA,
            how="left",
            on=["barcode_id"],
            validate="many_to_one",
        )
        .merge(
            right=df_ref,
            how="left",
            on=["chip", "ref_id"],
            validate="many_to_one",
        )
    )

    df_treat = df_treat.merge(
        right=df_column,
        how="left",
        on=["chip", "ref_id"],
        validate="many_to_one",
    ).assign(mh=lambda df: df.apply(utils.get_mh, axis=1))

    df_treat.to_csv(
        root_dir / "analyze" / "treat" / "annote" / "treat.csv",
        index=False,
        na_rep="NA",
    )

    df_control = df_control.merge(
        right=df_column,
        how="left",
        on=["chip", "ref_id"],
        validate="many_to_one",
    ).assign(mh=lambda df: df.apply(utils.get_mh, axis=1))

    df_control.to_csv(
        root_dir / "analyze" / "control" / "annote" / "control.csv",
        index=False,
        na_rep="NA",
    )



import os
import pathlib
import subprocess
from importlib import resources

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from . import utils


def get_file_pairs(data_dir: os.PathLike) -> list[dict[str, os.PathLike]]:
    data_dir = pathlib.Path(data_dir)
    file_pairs = []
    for file in os.listdir(data_dir):
        if not file.endswith(".R2.fq.gz"):
            continue
        stem = file.replace(".R2.fq.gz", "")
        file_pairs.append(
            {"R1": data_dir / f"{stem}.fq.gz", "R2": data_dir / f"{stem}.R2.fq.gz"}
        )
    return file_pairs


def remove_duplicates(data_dir: os.PathLike, root_dir: os.PathLike):
    file_pairs = get_file_pairs(data_dir)
    root_dir = pathlib.Path(os.fspath(root_dir))
    os.makedirs(root_dir / "unique", exist_ok=True)
    for file_pair in file_pairs:
        R1 = file_pair["R1"]
        R2 = file_pair["R2"]
        stem = R1.name.replace(".fq.gz", "")
        with open(root_dir / "unique" / f"{stem}.unique", "w") as fd:
            subprocess.run(
                args=[
                    "removeDuplicates.sh",
                    R1.as_posix(),
                    R2.as_posix(),
                ],
                stdout=fd,
            )


def agg(df: pd.DataFrame) -> pd.DataFrame:
    return (
        df.groupby(
            [
                "R1_primer",
                "G",
                "R1_sgRNA",
                "R1_scaffold_prefix",
                "R1_tail",
                "R2_primer",
                "barcode_head",
                "barcode",
                "CTG_target_prefix",
                "R2_sgRNA",
                "pam",
                "target_suffix",
                "C",
                "R2_scaffold_prefix",
                "R2_tail",
            ]
        )
        .agg(
            R1_primer_score=pd.NamedAgg(column="R1_primer_score", aggfunc="first"),
            R1_scaffold_prefix_score=pd.NamedAgg(
                column="R1_scaffold_prefix_score", aggfunc="first"
            ),
            R2_primer_score=pd.NamedAgg(column="R2_primer_score", aggfunc="first"),
            R2_scaffold_prefix_score=pd.NamedAgg(
                column="R2_scaffold_prefix_score", aggfunc="first"
            ),
            R2_sgRNA_score=pd.NamedAgg(column="R2_sgRNA_score", aggfunc="first"),
            barcode_score=pd.NamedAgg(column="barcode_score", aggfunc="first"),
            R1_sgRNA_bowtie2_score=pd.NamedAgg(
                column="R1_sgRNA_bowtie2_score", aggfunc="first"
            ),
            R2_sgRNA_bowtie2_score=pd.NamedAgg(
                column="R2_sgRNA_bowtie2_score", aggfunc="first"
            ),
            barcode_id=pd.NamedAgg(column="barcode_id", aggfunc="first"),
            R1_sgRNA_id=pd.NamedAgg(column="R1_sgRNA_id", aggfunc="first"),
            R2_sgRNA_id=pd.NamedAgg(column="R2_sgRNA_id", aggfunc="first"),
            count=pd.NamedAgg(column="count", aggfunc="sum"),
        )
        .sort_values("count", ascending=False)
        .reset_index()
    )


def stat(df: pd.DataFrame, save_dir: os.PathLike):
    save_dir = pathlib.Path(os.fspath(save_dir))
    os.makedirs(save_dir, exist_ok=True)
    for column in [
        "R1_primer",
        "R1_sgRNA",
        "R1_scaffold_prefix",
        "R1_tail",
        "R2_primer",
        "barcode_head",
        "barcode",
        "CTG_target_prefix",
        "R2_sgRNA",
        "pam",
        "target_suffix",
        "R2_scaffold_prefix",
        "R2_tail",
    ]:
        df_stat = (
            df[[column, "count"]]
            .assign(**{f"{column}_length": lambda df: df[column].str.len()})
            .groupby(f"{column}_length")["count"]
            .sum()
            .reset_index()
        )
        df_stat.to_csv(save_dir / f"{column}_length.csv", index=False)

    for column in [
        "G",
        "C",
        "pam_tail",
        "R1_primer_score",
        "R1_scaffold_prefix_score",
        "R2_primer_score",
        "R2_sgRNA_score",
        "R2_scaffold_prefix_score",
        "barcode_score",
        "R1_sgRNA_bowtie2_score",
        "R2_sgRNA_bowtie2_score",
        "barcode_id",
        "R1_sgRNA_id",
        "R2_sgRNA_id",
    ]:
        if column == "pam_tail":
            df = df.assign(pam_tail=lambda df: df["pam"].str.slice(start=-2))
        df_stat = df[[column, "count"]].groupby(column)["count"].sum().reset_index()
        df_stat.to_csv(save_dir / f"{column}.csv", index=False)

    df_stat = (
        df[["count"]]
        .groupby("count")
        .size()
        .reset_index()
        .rename(columns={"count": "count_full", 0: "count"})
    )
    df_stat.to_csv(save_dir / f"count_full.csv", index=False)

    df_stat = (
        df[["count"]].clip(upper=300).groupby("count").size().reset_index()
    ).rename(columns={"count": "count_small", 0: "count"})
    df_stat.to_csv(save_dir / f"count_small.csv", index=False)


def draw(save_dirs: list[os.PathLike], summary_dir: os.PathLike):
    save_dirs = [pathlib.Path(os.fspath(save_dir)) for save_dir in save_dirs]
    summary_dir = pathlib.Path(os.fspath(summary_dir))
    os.makedirs(summary_dir, exist_ok=True)
    for csv_file in os.listdir(save_dirs[0]):
        if not csv_file.endswith(".csv"):
            continue
        df_stat = pd.concat(
            [pd.read_csv(save_dir / csv_file, header=0) for save_dir in save_dirs]
        ).reset_index(drop=True)
        name = (
            df_stat.columns[0] if df_stat.columns[0] != "count" else df_stat.columns[1]
        )
        if name in [
            "R1_primer_length",
            "R2_primer_length",
            "barcode_id",
            "R1_sgRNA_id",
            "R2_sgRNA_id",
            "count_full",
            "count_small",
        ]:
            logy = True
        else:
            logy = False

        if name in ["G", "C", "pam_tail"]:
            df_stat.groupby(name)["count"].sum().plot.bar(
                logy=logy
            ).get_figure().savefig(summary_dir / f"{pathlib.Path(csv_file).stem}.pdf")
        elif name in ["barcode_id", "R1_sgRNA_id", "R2_sgRNA_id"]:
            df_stat.groupby(name)["count"].sum().reset_index().plot.scatter(
                x=name, y="count", logy=logy
            ).get_figure().savefig(summary_dir / f"{pathlib.Path(csv_file).stem}.pdf")
        elif name.endswith("_length") or name in ["count_full", "count_small"]:
            bins = 150 if name.endswith("_length") else 300
            df_stat[name].clip(upper=bins).plot.hist(
                bins=np.linspace(0, bins + 1, bins + 2),
                weights=df_stat["count"],
                logy=logy,
            ).get_figure().savefig(summary_dir / f"{pathlib.Path(csv_file).stem}.pdf")
        elif name.endswith("_score"):
            bins = 300
            df_stat[name].plot.hist(
                bins=bins, weights=df_stat["count"], logy=logy
            ).get_figure().savefig(summary_dir / f"{pathlib.Path(csv_file).stem}.pdf")
        else:
            raise ValueError("Unknown column")
        plt.close("all")


def collect_control(
    root_dir: os.PathLike,
):
    root_dir = pathlib.Path(os.fspath(root_dir))
    os.makedirs(root_dir / "control" / "full", exist_ok=True)
    for chip in ["a1", "a2", "a3", "g1n", "g2n", "g3n"]:
        df_controls = []
        for parse_file in os.listdir(root_dir / "parse" / "sgRNA"):
            if (
                utils.infer_cas(parse_file) != "control"
                or utils.infer_chip(parse_file) != chip
            ):
                continue
            df_controls.append(
                pd.read_csv(
                    root_dir / "parse" / "sgRNA" / parse_file,
                    sep="\t",
                    header=0,
                    keep_default_na=False,
                ).drop(columns=["R1_barcode", "R2_barcode"])
            )

        df_controls = pd.concat(df_controls)
        df_controls = agg(df_controls)
        df_controls.to_feather(root_dir / "control" / "full" / f"{chip}.feather")


def stat_control(root_dir: os.PathLike):
    root_dir = pathlib.Path(os.fspath(root_dir))
    for chip in ["a1", "a2", "a3", "g1n", "g2n", "g3n"]:
        save_dir = root_dir / "figures" / "align" / "stat_control" / chip
        df_control = pd.read_feather(root_dir / "control" / "full" / f"{chip}.feather")
        stat(df=df_control, save_dir=save_dir)
        draw(save_dirs=[save_dir], summary_dir=save_dir)
        del df_control


def filter_nofunc_control(
    root_dir: os.PathLike,
    range_R1_primer_length: list[int],
    range_R1_sgRNA_length: list[int],
    min_R1_scaffold_prefix_length: int,
    max_R1_tail_length: int,
    range_R2_primer_length: list[int],
    range_barcode_head_length: int,
    range_barcode_length: int,
    range_CTG_target_prefix_length: int,
    range_R2_sgRNA_length: list[int],
    range_pam_length: int,
    range_target_suffix_length: list[int],
    min_R2_scaffold_prefix_length: int,
    max_R2_tail_length: int,
    min_R1_primer_score: int,
    min_R1_scaffold_prefix_score: int,
    min_R2_primer_score: int,
    min_R2_sgRNA_score: int,
    min_R2_scaffold_prefix_score: int,
    min_barcode_score: int,
    min_R1_sgRNA_bowtie2_score: int,
    min_R2_sgRNA_bowtie2_score: int,
    G: list[str],
    C: list[str],
    a_pam_tail: list[str],
    g_pam_tail: list[str],
    min_count: int,
):
    root_dir = pathlib.Path(os.fspath(root_dir))
    os.makedirs(root_dir / "control" / "func", exist_ok=True)
    for chip in ["a1", "a2", "a3", "g1n", "g2n", "g3n"]:
        save_dir = root_dir / "figures" / "align" / "filter_nofunc_control" / chip
        os.makedirs(save_dir, exist_ok=True)
        df_stat = pd.DataFrame(columns=["row_num", "count"], index=["full", "filter"])

        df_control = pd.read_feather(root_dir / "control" / "full" / f"{chip}.feather")
        df_stat.loc["full", "row_num"] = df_control.shape[0]
        df_stat.loc["full", "count"] = df_control["count"].sum()

        pam_tail = a_pam_tail if chip in ["a1", "a2", "a3"] else g_pam_tail
        df_control.query(
            """
                @range_R1_primer_length[0] <= R1_primer.str.len() <= @range_R1_primer_length[1] and \
                @range_R1_sgRNA_length[0] <= R1_sgRNA.str.len() <= @range_R1_sgRNA_length[1] and \
                R1_scaffold_prefix.str.len() >= @min_R1_scaffold_prefix_length and \
                R1_tail.str.len() <= @max_R1_tail_length and \
                @range_R2_primer_length[0] <= R2_primer.str.len() <= @range_R2_primer_length[1] and \
                @range_barcode_head_length[0] <= barcode_head.str.len() <= @range_barcode_head_length[1] and \
                @range_barcode_length[0] <= barcode.str.len() <= @range_barcode_length[1] and \
                @range_CTG_target_prefix_length[0] <= CTG_target_prefix.str.len() <= @range_CTG_target_prefix_length[1] and \
                @range_R2_sgRNA_length[0] <= R2_sgRNA.str.len() <= @range_R2_sgRNA_length[1] and \
                @range_pam_length[0] <= pam.str.len() <= @range_pam_length[1] and \
                @range_target_suffix_length[0] <= target_suffix.str.len() <= @range_target_suffix_length[1] and \
                R2_scaffold_prefix.str.len() >= @min_R2_scaffold_prefix_length and \
                R2_tail.str.len() <= @max_R2_tail_length and \
                R1_primer_score >= @min_R1_primer_score and \
                R1_scaffold_prefix_score >= @min_R1_scaffold_prefix_score and \
                R2_primer_score >= @min_R2_primer_score and \
                R2_sgRNA_score >= @min_R2_sgRNA_score and \
                R2_scaffold_prefix_score >= @min_R2_scaffold_prefix_score and \
                barcode_score >= @min_barcode_score and \
                R1_sgRNA_bowtie2_score >= @min_R1_sgRNA_bowtie2_score and \
                R2_sgRNA_bowtie2_score >= @min_R2_sgRNA_bowtie2_score and \
                G.isin(@G) and \
                C.isin(@C) and \
                pam.str.slice(start=-2).isin(@pam_tail) and \
                count >= @min_count and \
                barcode_id >= 0 and \
                barcode_id == R1_sgRNA_id == R2_sgRNA_id
            """,
            inplace=True,
        )
        df_control.reset_index(drop=True, inplace=True)
        df_stat.loc["filter", "row_num"] = df_control.shape[0]
        df_stat.loc["filter", "count"] = df_control["count"].sum()

        df_stat.to_csv(save_dir / "stat.csv")
        df_stat["row_num"].plot.bar(logy=True).get_figure().savefig(
            save_dir / "row_num.pdf"
        )
        plt.close("all")
        df_stat["count"].plot.bar(logy=True).get_figure().savefig(
            save_dir / "count.pdf"
        )
        plt.close("all")

        df_control.to_feather(root_dir / "control" / "func" / f"{chip}.feather")


def cluster_func_control_by_mutant(
    root_dir: os.PathLike, ext: int, plasmid_file: os.PathLike | None
):
    def infer_indel(row: pd.Series) -> pd.Series:
        ref_end1 = row["ref_end1"]
        ref_start2 = row["ref_start2"]
        random_insertion = row["random_insertion"]
        cut1 = row["cut1"]
        cut2 = row["cut2"]
        if ref_end1 - cut1 == ref_start2 - cut2:
            if random_insertion == "":
                indel_type = "wt"
                pos = 0
                size = 0
            else:
                indel_type = "ins"
                pos = 0
                size = len(random_insertion)
        elif ref_end1 - cut1 < ref_start2 - cut2:
            pos = ref_end1 - cut1
            size = ref_start2 - cut2 - ref_end1 + cut1 + len(random_insertion)
            if random_insertion == "":
                indel_type = "del"
            else:
                indel_type = "indel"
        else:
            indel_type = "ins"
            pos = ref_start2 - cut2
            size = ref_end1 - cut1 - ref_start2 + cut2 + len(random_insertion)

        return f"{indel_type}_{pos}_{size}".replace("-", "m")

    root_dir = pathlib.Path(os.fspath(root_dir))
    os.makedirs(root_dir / "control" / "cluster", exist_ok=True)
    for chip in ["a1", "a2", "a3", "g1n", "g2n", "g3n"]:
        if plasmid_file is None:
            if chip in ["a1", "a2", "a3"]:
                plasmid_file_default = (
                    "final_hgsgrna_libb_all_0811_NAA_scaffold_nbt.csv"
                )
            else:
                plasmid_file_default = "final_hgsgrna_libb_all_0811-NGG.csv"
            with resources.as_file(
                resources.files() / "plasmids" / plasmid_file_default
            ) as pf:
                df_plasmid = pd.read_csv(pf, header=0)
        else:
            df_plasmid = pd.read_csv(plasmid_file, header=0)
        df_plasmid = df_plasmid.assign(
            ref2=lambda df: (
                "AAATAAGGCTAGTCCGTTATCAACTTGAAAAAGTGGCACCGAGTCGGTGCTTTTTTG"
                + df["Target sequence"].str.slice(stop=22 + ext)
            ).map(utils.rev_comp),
            ref1=lambda df: (
                df["Target sequence"].str.slice(start=22 - ext)
                + "CAG"
                + df["Barcode2"]
                + df["Primer binding sites 20nt"]
            ).map(utils.rev_comp),
            cut=lambda df: df["ref1"].str.len() - ext,
            ext=ext,
        ).reset_index(names="barcode_id")

        df_control = pd.read_feather(root_dir / "control" / "func" / f"{chip}.feather")
        df_control = df_control.assign(
            query=lambda df: df["R2_primer"]
            + df["barcode_head"]
            + df["barcode"]
            + df["CTG_target_prefix"]
            + df["R2_sgRNA"]
            + df["pam"]
            + df["target_suffix"]
            + df["C"]
            + df["R2_scaffold_prefix"]
            + df["R2_tail"],
        )[["barcode_id", "query", "count"]].merge(
            right=df_plasmid[["barcode_id", "ref1", "ref2", "cut", "ext"]],
            how="left",
            on=["barcode_id"],
            validate="many_to_one",
        )

        df_alg = utils.call_rearr(
            df_control["ref1"],
            df_control["ref2"],
            df_control["cut"],
            df_control["ext"],
            df_control["query"],
            df_control["count"],
        )

        df_control = (
            df_control.assign(
                ref_end1=df_alg["ref_end1"],
                ref_start2=df_alg["ref_start2"],
                random_insertion=df_alg["random_insertion"],
                cut1=lambda df: df["cut"],
                cut2=lambda df: df["cut1"] + 2 * ext,
            )
            .groupby(
                [
                    "barcode_id",
                    "ref_end1",
                    "ref_start2",
                    "random_insertion",
                    "cut1",
                    "cut2",
                    "ref1",
                    "ref2",
                ]
            )["count"]
            .sum()
            .sort_values()
            .reset_index()
            .assign(
                percentage=lambda df: df.groupby("barcode_id")["count"].transform(
                    "cumsum"
                )
                / df.groupby("barcode_id")["count"].transform("sum"),
                rank=lambda df: df.groupby("barcode_id")["count"].rank(ascending=False),
            )
        )

        # count_wt
        df_control = (
            df_control.merge(
                right=df_control.query(
                    """
                    ref_end1 == cut1 and \
                    ref_start2 == cut2 and \
                    random_insertion == ""
                """
                )[["barcode_id", "count"]].rename(columns={"count": "count_wt"}),
                how="left",
                on=["barcode_id"],
                validate="many_to_one",
            )
            .assign(**{"count_wt": lambda df: df["count_wt"].fillna(0)})
            .astype({"count_wt": int})
        )

        # annote indel type
        df_control = df_control.assign(
            indel_type=lambda df: df.apply(infer_indel, axis=1)
        )

        all_counts = ["count_wt"]
        for it in ["del", "ins"]:
            for size in range(1, 4):
                for pos in range(-ext - size, ext + 1):
                    indel_type = f"{it}_{pos}_{size}".replace("-", "m")
                    df_control = df_control.merge(
                        right=df_control.query("indel_type == @indel_type")
                        .groupby("barcode_id")["count"]
                        .sum()
                        .rename(f"count_{indel_type}")
                        .reset_index(),
                        how="left",
                        on=["barcode_id"],
                        validate="many_to_one",
                    ).assign(
                        **{
                            f"count_{indel_type}": lambda df: df[
                                f"count_{indel_type}"
                            ].fillna(0)
                        }
                    )
                    all_counts.append(f"count_{indel_type}")

                df_control = df_control.copy()

        # count_tot, first, second, first_all, second_all
        df_control = df_control.assign(
            count_tot=lambda df: df.groupby("barcode_id")["count"].transform("sum"),
            first=lambda df: df[all_counts].max(axis=1),
            second=lambda df: df[all_counts].apply(
                lambda row: row.nlargest(2).min(), axis=1
            ),
            first_all=lambda df: df.groupby("barcode_id")["count"].transform("max"),
            second_all=lambda df: df.groupby("barcode_id")["count"].transform(
                lambda se: se.nlargest(2).min()
            ),
        )
        df_control["second_all"] = df_control["second_all"].where(
            df_control.groupby("barcode_id").transform("size") > 1, 0
        )

        # mutant_type_num
        df_control = df_control.assign(
            mutant_type_num=lambda df: df.groupby("barcode_id").transform("size")
        )

        df_control.to_feather(root_dir / "control" / "cluster" / f"{chip}.feather")


def stat_func_control(root_dir: os.PathLike, ext: int):
    root_dir = pathlib.Path(os.fspath(root_dir))
    for chip in ["a1", "a2", "a3", "g1n", "g2n", "g3n"]:
        save_dir = root_dir / "figures" / "align" / "stat_func_control" / chip
        os.makedirs(save_dir, exist_ok=True)
        df_control = pd.read_feather(
            root_dir / "control" / "cluster" / f"{chip}.feather"
        )

        # number of mutant types per barcode_id
        df_control.groupby("barcode_id")["mutant_type_num"].first().clip(
            upper=100
        ).plot.hist(bins=np.linspace(0, 101, 102)).get_figure().savefig(
            save_dir / "mutant_type_num.pdf"
        )
        plt.close("all")

        # up_del_size
        df_stat = (
            df_control.assign(up_del_size=lambda df: utils.up_del_size(df))
            .groupby("up_del_size")["count"]
            .sum()
            .reset_index()
        )
        df_stat["up_del_size"].plot.hist(
            bins=np.linspace(0, 101, 102), weights=df_stat["count"]
        ).get_figure().savefig(save_dir / "up_del_size.pdf")
        plt.close("all")

        # down_del_size
        df_stat = (
            df_control.assign(down_del_size=lambda df: utils.down_del_size(df))
            .groupby("down_del_size")["count"]
            .sum()
            .reset_index()
        )
        df_stat["down_del_size"].plot.hist(
            bins=np.linspace(0, 101, 102), weights=df_stat["count"]
        ).get_figure().savefig(save_dir / "down_del_size.pdf")
        plt.close("all")

        # rand_ins_size
        df_stat = (
            df_control.assign(rand_ins_size=lambda df: utils.rand_ins_size(df))
            .groupby("rand_ins_size")["count"]
            .sum()
            .reset_index()
        )
        df_stat["rand_ins_size"].plot.hist(
            bins=np.linspace(0, 101, 102), weights=df_stat["count"]
        ).get_figure().savefig(save_dir / "rand_ins_size.pdf")
        plt.close("all")

        # count
        df_control["count"].clip(upper=300).plot.hist(
            bins=np.linspace(0, 301, 302),
        ).get_figure().savefig(save_dir / "count.pdf")
        plt.close("all")

        # count_tot
        df_control.groupby("barcode_id")["count_tot"].first().clip(upper=300).plot.hist(
            bins=np.linspace(0, 301, 302),
        ).get_figure().savefig(save_dir / "count_tot.pdf")
        plt.close("all")

        # freq_wt
        df_control.assign(freq_wt=lambda df: df["count_wt"] / df["count_tot"]).groupby(
            "barcode_id"
        )["freq_wt"].first().plot.hist(bins=100, logy=True).get_figure().savefig(
            save_dir / "freq_wt.pdf"
        )
        plt.close("all")

        # second_rel_first
        df_control.assign(
            second_rel_first=lambda df: df["second"] / (df["first"] + 1e-6)
        ).groupby("barcode_id")["second_rel_first"].first().plot.hist(
            bins=100, logy=True
        ).get_figure().savefig(
            save_dir / "second_rel_first.pdf"
        )
        plt.close("all")

        # second_all_rel_first_all
        df_control.assign(
            second_all_rel_first_all=lambda df: df["second_all"]
            / (df["first_all"] + 1e-6)
        ).groupby("barcode_id")["second_all_rel_first_all"].first().plot.hist(
            bins=100, logy=True
        ).get_figure().savefig(
            save_dir / "second_all_rel_first_all.pdf"
        )
        plt.close("all")

        # type_max
        all_counts = ["count_wt"]
        for it in ["del", "ins"]:
            for size in range(1, 4):
                for pos in range(-ext - size, ext + 1):
                    indel_type = f"{it}_{pos}_{size}".replace("-", "m")
                    all_counts.append(f"count_{indel_type}")
        df_stat = (
            df_control.assign(type_max=lambda df: df[all_counts].idxmax(axis=1))
            .groupby("barcode_id")
            .agg(
                count_tot=pd.NamedAgg(column="count", aggfunc="sum"),
                type_max=pd.NamedAgg(column="type_max", aggfunc="first"),
            )
            .reset_index()
            .groupby("type_max")
            .agg(
                num_type_max=pd.NamedAgg(column="type_max", aggfunc="size"),
                count_type_max=pd.NamedAgg(column="count_tot", aggfunc="sum"),
            )
        )
        df_stat["num_type_max"].plot.bar(logy=True).get_figure().savefig(
            save_dir / "num_type_max.pdf"
        )
        plt.close("all")
        df_stat["count_type_max"].plot.bar(logy=True).get_figure().savefig(
            save_dir / "count_type_max.pdf"
        )
        plt.close("all")


def filter_low_quality_barcode(
    root_dir: os.PathLike,
    max_mutant_type_num: int,
    min_count_tot: int,
    min_freq_wt: float,
    max_second_rel_first: float,
    max_second_all_rel_first_all: float,
):
    root_dir = pathlib.Path(os.fspath(root_dir))
    os.makedirs(root_dir / "control" / "hq_bar", exist_ok=True)
    for chip in ["a1", "a2", "a3", "g1n", "g2n", "g3n"]:
        save_dir = root_dir / "figures" / "align" / "filter_low_quality_barcode" / chip
        os.makedirs(save_dir, exist_ok=True)
        df_control = pd.read_feather(
            root_dir / "control" / "cluster" / f"{chip}.feather"
        )

        df_stat = pd.DataFrame(columns=["bar_num"], index=["full", "filter"])
        df_stat.loc["full", "bar_num"] = df_control["barcode_id"].unique().shape[0]

        df_control = df_control.query(
            """
                mutant_type_num <= @max_mutant_type_num and \
                (count_tot >= @min_count_tot or count_wt / count_tot >= @min_freq_wt) and \
                second / (first + 1e-6) <= @max_second_rel_first and \
                second_all / (first_all + 1e-6) <= @max_second_all_rel_first_all
            """
        ).reset_index(drop=True)

        df_stat.loc["filter", "bar_num"] = df_control["barcode_id"].unique().shape[0]
        df_stat["bar_num"].plot.bar().get_figure().savefig(save_dir / "bar_num.pdf")
        plt.close("all")

        df_control.to_feather(root_dir / "control" / "hq_bar" / f"{chip}.feather")


def filter_low_quality_mutant(
    root_dir: os.PathLike,
    min_percentage: float,
    max_rank: int,
    max_up_del_size: int,
    max_down_del_size: int,
    max_rand_ins_size: int,
    min_count: int,
):
    assert max_down_del_size <= 3, "pam is resected"
    root_dir = pathlib.Path(os.fspath(root_dir))
    os.makedirs(root_dir / "control" / "hq_mut", exist_ok=True)
    for chip in ["a1", "a2", "a3", "g1n", "g2n", "g3n"]:
        save_dir = root_dir / "figures" / "align" / "filter_low_quality_mutant" / chip
        os.makedirs(save_dir, exist_ok=True)
        df_control = pd.read_feather(
            root_dir / "control" / "hq_bar" / f"{chip}.feather"
        )

        df_stat = pd.DataFrame(columns=["mut_num"], index=["full", "filter"])
        df_stat.loc["full", "mut_num"] = df_control.shape[0]

        df_control = df_control.query(
            """
                percentage >= @min_percentage and \
                rank <= @max_rank and \
                cut1 - ref_end1 <= @max_up_del_size and \
                ref_start2 - cut2 <= @max_down_del_size and \
                random_insertion.str.len() <= @max_rand_ins_size and \
                count >= @min_count
            """
        ).reset_index(drop=True)

        df_stat.loc["filter", "mut_num"] = df_control.shape[0]
        df_stat["mut_num"].plot.bar().get_figure().savefig(save_dir / "mut_num.pdf")
        plt.close("all")

        df_control.groupby("barcode_id").size().rename("mutant_type_num").clip(
            upper=100
        ).plot.hist(bins=np.linspace(0, 101, 102)).get_figure().savefig(
            save_dir / "mutant_type_num.pdf"
        )
        plt.close("all")

        df_control.to_feather(root_dir / "control" / "hq_mut" / f"{chip}.feather")


def generate_reference(
    root_dir: os.PathLike,
    ext: int,
):
    def generate_reference_row(row: pd.Series, ext: int) -> pd.Series:
        ref = (
            row["ref1"][: row["ref_end1"]]
            + row["random_insertion"]
            + row["ref2"][row["ref_start2"] - len(row["ref1"]) :]
        )
        ref1 = row["ref1"][: row["ref_end1"]] + row["random_insertion"]
        if row["ref_start2"] > row["cut2"]:
            assert len(ref1) >= row["ref_start2"] - row["cut2"], "upstream too short"
            ref1 = ref1[: (row["cut2"] - row["ref_start2"])]
        elif row["ref_start2"] < row["cut2"]:
            ref1 = (
                ref1
                + row["ref2"][
                    row["ref_start2"]
                    - len(row["ref1"]) : row["cut2"]
                    - len(row["ref1"])
                ]
            )
        cut = len(ref1)
        ref1 = ref[: cut + ext]
        ref2 = ref[cut - ext :]
        return pd.Series(
            {
                "ref1": ref1,
                "ref2": ref2,
                "cut": cut,
            }
        )

    root_dir = pathlib.Path(root_dir)
    os.makedirs(root_dir / "ref", exist_ok=True)
    for chip in ["a1", "a2", "a3", "g1n", "g2n", "g3n"]:
        df_control = pd.read_feather(
            root_dir / "control" / "hq_mut" / f"{chip}.feather"
        )

        assert (
            df_control["ref_start2"] <= df_control["cut2"] + 3
        ).all(), "pam is mutated"

        df_ref = (
            df_control[
                ["ref_end1", "ref_start2", "random_insertion", "cut2", "ref1", "ref2"]
            ]
            .apply(lambda row, ext=ext: generate_reference_row(row, ext), axis=1)
            .assign(
                zero=0,
                ext=ext,
                ref2len=lambda df: df["ref2"].str.len(),
            )
        )

        df_ref[["zero", "ref1", "cut", "ext", "ref2", "ref2len"]].to_csv(
            root_dir / "ref" / f"{chip}.ref",
            sep="\t",
            header=False,
            index=False,
        )


def collect_treat(root_dir: os.PathLike):
    root_dir = pathlib.Path(os.fspath(root_dir))
    os.makedirs(root_dir / "treat" / "full", exist_ok=True)
    for parse_file in os.listdir(root_dir / "parse" / "sgRNA"):
        df_treat = pd.read_csv(
            root_dir / "parse" / "sgRNA" / parse_file,
            sep="\t",
            header=0,
            keep_default_na=False,
        ).drop(columns=["R1_barcode", "R2_barcode"])

        df_treat = agg(df_treat)
        df_treat.to_feather(
            root_dir / "treat" / "full" / f"{pathlib.Path(parse_file).stem}.feather"
        )


def stat_treat(root_dir: os.PathLike):
    root_dir = pathlib.Path(os.fspath(root_dir))
    save_dir = root_dir / "figures" / "align" / "stat_treat"
    save_dirs = []
    for treat_file in os.listdir(root_dir / "treat" / "full"):
        df_treat = pd.read_feather(root_dir / "treat" / "full" / treat_file)
        stat(
            df=df_treat,
            save_dir=save_dir / pathlib.Path(treat_file).stem,
        )
        save_dirs.append(save_dir / pathlib.Path(treat_file).stem)

    draw(save_dirs, summary_dir=save_dir)


def filter_treat(
    root_dir: os.PathLike,
    range_R1_primer_length: list[int],
    range_R1_sgRNA_length: list[int],
    # min_R1_scaffold_prefix_length: int,
    # max_R1_tail_length: int,
    range_R2_primer_length: list[int],
    # range_barcode_head_length: int,
    # range_barcode_length: int,
    # range_CTG_target_prefix_length: int,
    # range_R2_sgRNA_length: list[int],
    # range_target_suffix_length: list[int],
    # min_R2_scaffold_prefix_length: int,
    # max_R2_tail_length: int,
    # range_pam_length: int,
    min_R1_primer_score: int,
    # min_R1_scaffold_prefix_score: int,
    min_R2_primer_score: int,
    # min_R2_sgRNA_score: int,
    # min_R2_scaffold_prefix_score: int,
    # min_barcode_score: int,
    min_R1_sgRNA_bowtie2_score: int,
    # min_R2_sgRNA_bowtie2_score: int,
    G: list[str],
    # C: list[str],
    # a_pam_tail: list[str],
    # g_pam_tail: list[str],
    min_count: int,
):
    root_dir = pathlib.Path(os.fspath(root_dir))
    os.makedirs(root_dir / "treat" / "filter", exist_ok=True)
    for treat_file in os.listdir(root_dir / "treat" / "full"):
        save_dir = (
            root_dir
            / "figures"
            / "align"
            / "filter_treat"
            / pathlib.Path(treat_file).stem
        )
        os.makedirs(save_dir, exist_ok=True)

        df_treat = pd.read_feather(root_dir / "treat" / "full" / treat_file)

        df_stat = pd.DataFrame(columns=["row_num", "count"], index=["full", "filter"])
        df_stat.loc["full", "row_num"] = df_treat.shape[0]
        df_stat.loc["full", "count"] = df_treat["count"].sum()

        df_treat = df_treat.query(
            """
                @range_R1_primer_length[0] <= R1_primer.str.len() <= @range_R1_primer_length[1] and \
                @range_R1_sgRNA_length[0] <= R1_sgRNA.str.len() <= @range_R1_sgRNA_length[1] and \
                @range_R2_primer_length[0] <= R2_primer.str.len() <= @range_R2_primer_length[1] and \
                R1_primer_score >= @min_R1_primer_score and \
                R2_primer_score >= @min_R2_primer_score and \
                R1_sgRNA_bowtie2_score >= @min_R1_sgRNA_bowtie2_score and \
                G.isin(@G) and \
                count >= @min_count and \
                barcode_id >= 0 and \
                barcode_id == R1_sgRNA_id
            """
        ).reset_index(drop=True)
        df_stat.loc["filter", "row_num"] = df_treat.shape[0]
        df_stat.loc["filter", "count"] = df_treat["count"].sum()

        df_stat.to_csv(save_dir / "stat.csv", index=False)
        df_stat["row_num"].plot.bar().get_figure().savefig(save_dir / "row_num.pdf")
        plt.close("all")
        df_stat["count"].plot.bar().get_figure().savefig(save_dir / "count.pdf")
        plt.close("all")

        df_treat.to_feather(root_dir / "treat" / "filter" / treat_file)


def demultiplex(root_dir: os.PathLike):
    root_dir = pathlib.Path(os.fspath(root_dir))
    os.makedirs(root_dir / "query" / "found", exist_ok=True)
    os.makedirs(root_dir / "query" / "not_found", exist_ok=True)
    for treat_file in os.listdir(root_dir / "treat" / "filter"):
        save_dir = (
            root_dir
            / "figures"
            / "align"
            / "demultiplex"
            / pathlib.Path(treat_file).stem
        )
        os.makedirs(save_dir, exist_ok=True)

        chip = utils.infer_chip(treat_file)
        df_ref = pd.read_feather(root_dir / "control" / "hq_mut" / f"{chip}.feather")[
            ["barcode_id"]
        ].reset_index(names="ref_id")

        df_query = (
            pd.read_feather(root_dir / "treat" / "filter" / treat_file)
            .assign(
                query=lambda df: df["R2_primer"]
                + df["barcode_head"]
                + df["barcode"]
                + df["CTG_target_prefix"]
                + df["R2_sgRNA"]
                + df["pam"]
                + df["target_suffix"]
                + df["C"]
                + df["R2_scaffold_prefix"]
                + df["R2_tail"]
            )
            .groupby("query")
            .agg(
                barcode_id=pd.NamedAgg(column="barcode_id", aggfunc="first"),
                count=pd.NamedAgg(column="count", aggfunc="sum"),
            )
            .reset_index()
            .merge(
                right=df_ref[["barcode_id", "ref_id"]],
                how="left",
                on=["barcode_id"],
            )
            .assign(
                ref_id=lambda df: df["ref_id"].fillna(-1).astype(int),
                count_distri=lambda df: df["count"]
                / df.groupby("query").transform("size"),
            )
        )

        # Get query number/count distribution for each ref. Count of query with multiple refs are distributed to each ref evenly.
        df_query.groupby("ref_id").size().rename("number").to_csv(
            save_dir / "query_number_per_ref.csv"
        )
        df_query.groupby("ref_id")["count_distri"].sum().rename("count").to_csv(
            save_dir / "query_count_per_ref.csv"
        )

        df_query.query("ref_id == -1")[["query", "count"]].to_csv(
            root_dir / "query" / "not_found" / f"{pathlib.Path(treat_file).stem}.query",
            sep="\t",
            header=False,
            index=False,
        )
        df_query.query("ref_id != -1")[["query", "count", "ref_id"]].to_csv(
            root_dir / "query" / "found" / f"{pathlib.Path(treat_file).stem}.query",
            sep="\t",
            header=False,
            index=False,
        )

    summary_demultiplex(save_dir=root_dir / "figures" / "align" / "demultiplex")


def summary_demultiplex(save_dir: os.PathLike):
    save_dir = pathlib.Path(os.fspath(save_dir))
    df_query_number_per_refs = []
    df_query_count_per_refs = []
    for stem in os.listdir(save_dir):
        if not os.path.isdir(save_dir / stem):
            continue
        df_query_number_per_refs.append(
            pd.read_csv(save_dir / stem / "query_number_per_ref.csv", header=0).assign(
                stem=stem
            )
        )
        df_query_count_per_refs.append(
            pd.read_csv(save_dir / stem / "query_count_per_ref.csv", header=0).assign(
                stem=stem
            )
        )

    pd.concat(df_query_number_per_refs).groupby("ref_id")[
        "number"
    ].sum().reset_index().plot.scatter(
        x="ref_id", y="number", logy=True
    ).get_figure().savefig(
        save_dir / "query_number_per_ref.pdf"
    )
    plt.close("all")

    pd.concat(df_query_count_per_refs).groupby("ref_id")[
        "count"
    ].sum().reset_index().plot.scatter(
        x="ref_id", y="count", logy=True
    ).get_figure().savefig(
        save_dir / "query_count_per_ref.pdf"
    )
    plt.close("all")



import subprocess
import sys
from importlib import resources


def parse():
    with resources.as_file(resources.files() / "parse.sh") as pf:
        subprocess.run(args=["bash", pf.as_posix(), sys.argv[1]])


def align():
    with resources.as_file(resources.files() / "align.sh") as pf:
        subprocess.run(args=["bash", pf.as_posix(), sys.argv[1]])



import os
import pathlib
import subprocess
import sys
from importlib import resources

import pandas as pd
import pysam
from Bio import Align, Seq

from . import utils


def get_embedding_aligner() -> Align.PairwiseAligner:
    aligner = Align.PairwiseAligner(scoring="blastn")
    aligner.mode = "global"
    aligner.open_left_deletion_score = 0
    aligner.extend_left_deletion_score = 0
    aligner.open_right_deletion_score = 0
    aligner.extend_right_deletion_score = 0

    return aligner


def get_suffix_aligner() -> Align.PairwiseAligner:
    aligner = Align.PairwiseAligner(scoring="blastn")
    aligner.mode = "global"
    aligner.open_left_deletion_score = 0
    aligner.extend_left_deletion_score = 0
    aligner.open_right_deletion_score = 0
    aligner.extend_right_deletion_score = 0
    aligner.open_right_insertion_score = 0
    aligner.extend_right_insertion_score = 0

    return aligner


def fuzzy_split(
    target: str, delimiter: str, aligner: Align.PairwiseAligner
) -> tuple[str | int]:
    if target == "":
        return "", "", "", 0
    if delimiter == "":
        return "", "", target, 0
    segs = target.split(delimiter, maxsplit=1)
    if len(segs) == 2:
        score = aligner.substitution_matrix[0, 0] * len(delimiter)
        return (
            segs[0],
            delimiter,
            segs[1],
            score,
        )
    alignment = aligner.align(target, delimiter)[0]
    if alignment.aligned.shape[1] == 0:
        start = 0
        end = 0
    else:
        start = alignment.aligned[0][0][0]
        end = alignment.aligned[0][-1][-1]
    return (
        alignment.target[:start],
        alignment.target[start:end],
        alignment.target[end:],
        alignment.score,
    )


def parse_read(
    read: str,
    primer_temp: str,
    scaffold_temps: list[str],
    embedding_aligner: Align.PairwiseAligner,
    suffix_aligner: Align.PairwiseAligner,
) -> tuple[str | list[str]]:
    barcode, primer, remain, primer_score = fuzzy_split(
        target=read, delimiter=primer_temp, aligner=embedding_aligner
    )
    scaffold_prefix_score = -float("inf")
    for idx, scaffold_temp in enumerate(scaffold_temps):
        new = fuzzy_split(
            target=remain, delimiter=scaffold_temp, aligner=suffix_aligner
        )
        if new[-1] > scaffold_prefix_score:
            variant, scaffold_prefix, tail, scaffold_prefix_score = new
            used_scaffold_idxs = [idx]
        elif new[-1] == scaffold_prefix_score:
            used_scaffold_idxs.append(idx)

    return (
        barcode,
        primer,
        variant,
        scaffold_prefix,
        tail,
        primer_score,
        scaffold_prefix_score,
        used_scaffold_idxs,
    )


def parse_R1(R1_variant: str) -> tuple[str]:
    G = R1_variant[:1]
    R1_sgRNA = R1_variant[1:]
    return G, R1_sgRNA


def parse_R2(
    R2_variant: str, R1_sgRNA: str, embedding_aliger: Align.PairwiseAligner
) -> tuple[str]:
    barcode_CTG_target = R2_variant[:-1]
    C = R2_variant[-1:]
    barcode_CTG_target_prefix, R2_sgRNA, pam_target_suffix, R2_sgRNA_score = (
        fuzzy_split(
            target=barcode_CTG_target, delimiter=R1_sgRNA, aligner=embedding_aliger
        )
    )
    pam = pam_target_suffix[:3]
    target_suffix = pam_target_suffix[3:]

    return barcode_CTG_target_prefix, R2_sgRNA, pam, target_suffix, C, R2_sgRNA_score


def main():
    unique_file = pathlib.Path(sys.argv[1])
    parse_file = (
        unique_file.parent.parent / "parse" / "nobar" / f"{unique_file.stem}.parse"
    )
    os.makedirs(parse_file.parent, exist_ok=True)
    chip = utils.infer_chip(unique_file)
    embedding_aligner = get_embedding_aligner()
    suffix_aligner = get_suffix_aligner()
    R1_primer_temp = utils.p5primer()
    R2_primer_temp = utils.p7primer()
    R1_scaffold_temps = [utils.scaffold(chip), utils.scaffold_alt(chip)]
    R2_scaffold_temps = [utils.RCscaffold(chip), utils.RCscaffold_alt(chip)]
    with open(unique_file, "r") as rd, open(parse_file, "w") as wd:
        wd.write(
            "R1_barcode\tR1_primer\tG\tR1_sgRNA\tR1_scaffold_prefix\tR1_tail\tR1_primer_score\tR1_scaffold_prefix_score\tR2_barcode\tR2_primer\tbarcode_CTG_target_prefix\tR2_sgRNA\tpam\ttarget_suffix\tC\tR2_scaffold_prefix\tR2_tail\tR2_primer_score\tR2_scaffold_prefix_score\tR2_sgRNA_score\tcount\n"
        )
        for line in rd:
            R1, R2, count = line.split()
            (
                R1_barcode,
                R1_primer,
                R1_variant,
                R1_scaffold_prefix,
                R1_tail,
                R1_primer_score,
                R1_scaffold_prefix_score,
                used_scaffold_idxs,
            ) = parse_read(
                read=R1,
                primer_temp=R1_primer_temp,
                scaffold_temps=R1_scaffold_temps,
                embedding_aligner=embedding_aligner,
                suffix_aligner=suffix_aligner,
            )
            G, R1_sgRNA = parse_R1(R1_variant)
            (
                R2_barcode,
                R2_primer,
                R2_variant,
                R2_scaffold_prefix,
                R2_tail,
                R2_primer_score,
                R2_scaffold_prefix_score,
                _,
            ) = parse_read(
                read=R2,
                primer_temp=R2_primer_temp,
                scaffold_temps=[R2_scaffold_temps[idx] for idx in used_scaffold_idxs],
                embedding_aligner=embedding_aligner,
                suffix_aligner=suffix_aligner,
            )
            (
                barcode_CTG_target_prefix,
                R2_sgRNA,
                pam,
                target_suffix,
                C,
                R2_sgRNA_score,
            ) = parse_R2(R2_variant, R1_sgRNA, embedding_aligner)

            wd.write(
                f"{R1_barcode}\t{R1_primer}\t{G}\t{R1_sgRNA}\t{R1_scaffold_prefix}\t{R1_tail}\t{R1_primer_score}\t{R1_scaffold_prefix_score}\t{R2_barcode}\t{R2_primer}\t{barcode_CTG_target_prefix}\t{R2_sgRNA}\t{pam}\t{target_suffix}\t{C}\t{R2_scaffold_prefix}\t{R2_tail}\t{R2_primer_score}\t{R2_scaffold_prefix_score}\t{R2_sgRNA_score}\t{count}\n"
            )


def build_barcode(root_dir: os.PathLike, plasmid_file: os.PathLike | None):
    root_dir = pathlib.Path(os.fspath(root_dir))
    os.makedirs(root_dir / "barcode" / "index", exist_ok=True)
    if plasmid_file is None:
        with resources.as_file(
            resources.files().parent
            / "plasmids"
            / "plasmids/final_hgsgrna_libb_all_0811_NAA_scaffold_nbt.csv"
        ) as pf:
            df_plasmid = pd.read_csv(pf, header=0)
    else:
        df_plasmid = pd.read_csv(plasmid_file, header=0)

    with open(root_dir / "barcode" / "index" / "barcode.fa", "w") as fd:
        for i, barcode in enumerate(df_plasmid["Barcode2"]):
            barcode = str(Seq.Seq(barcode).reverse_complement())
            fd.write(f">b{i}\n{barcode}\n")

        subprocess.run(
            args=[
                "bowtie2-build",
                (root_dir / "barcode" / "index" / "barcode.fa").as_posix(),
                (root_dir / "barcode" / "index" / "barcode").as_posix(),
            ]
        )


def prepare_barcode_CTG_target_prefix(root_dir: os.PathLike):
    root_dir = pathlib.Path(os.fspath(root_dir))
    os.makedirs(root_dir / "barcode" / "read", exist_ok=True)
    for parse_file in os.listdir(root_dir / "parse" / "nobar"):
        df = pd.read_csv(
            root_dir / "parse" / "nobar" / parse_file,
            sep="\t",
            header=0,
            usecols=["barcode_CTG_target_prefix"],
            keep_default_na=False,
        )
        df["barcode_CTG_target_prefix"] = df["barcode_CTG_target_prefix"].where(
            df["barcode_CTG_target_prefix"] != "", "N"
        )
        df.to_csv(
            root_dir / "barcode" / "read" / f"{pathlib.Path(parse_file).stem}.csv",
            header=False,
            index=False,
        )


def map_barcode(root_dir: os.PathLike):
    root_dir = pathlib.Path(os.fspath(root_dir))
    os.makedirs(root_dir / "barcode" / "align", exist_ok=True)
    for csv_file in os.listdir(root_dir / "barcode" / "read"):
        subprocess.run(
            args=[
                "bowtie2",
                "--quiet",
                "-p",
                "24",
                "--mm",
                "--norc",
                "--local",
                "-L",
                "7",
                "--ma",
                "1",
                "--mp",
                "2,2",
                "--rdg",
                "3,1",
                "--rfg",
                "3,1",
                "--score-min",
                "C,1",
                "-r",
                "-x",
                (root_dir / "barcode" / "index" / "barcode").as_posix(),
                "-U",
                (root_dir / "barcode" / "read" / csv_file).as_posix(),
                "-S",
                (
                    root_dir
                    / "barcode"
                    / "align"
                    / f"{pathlib.Path(csv_file).stem}.sam"
                ).as_posix(),
            ],
        )


def parse_barcode(root_dir: os.PathLike):
    root_dir = pathlib.Path(os.fspath(root_dir))
    os.makedirs(root_dir / "parse" / "bar", exist_ok=True)
    for parse_file in os.listdir(root_dir / "parse" / "nobar"):
        sam_file = f"{pathlib.Path(parse_file).stem}.sam"
        sam = pysam.AlignmentFile(root_dir / "barcode" / "align" / sam_file)
        qnames = []
        barcode_heads = []
        barcodes = []
        CTG_target_prefixs = []
        barcode_ids = []
        barcode_scores = []
        for align in sam.fetch():
            qnames.append(int(align.qname))
            if (align.flag // 4) % 2 == 1:
                barcode_heads.append("")
                barcodes.append("")
                CTG_target_prefixs.append(align.seq if align.seq != "N" else "")
                barcode_ids.append(-1)
                barcode_scores.append(0)
            else:
                barcode_heads.append(align.seq[: align.qstart])
                barcodes.append(align.seq[align.qstart : align.qend])
                CTG_target_prefixs.append(align.seq[align.qend :])
                barcode_ids.append(align.rname)
                barcode_scores.append(align.get_tag("AS"))
        df_sam = pd.DataFrame(
            {
                "qname": qnames,
                "barcode_head": barcode_heads,
                "barcode": barcodes,
                "CTG_target_prefix": CTG_target_prefixs,
                "barcode_id": barcode_ids,
                "barcode_score": barcode_scores,
            }
        )

        df_parse = (
            pd.read_csv(
                root_dir / "parse" / "nobar" / parse_file,
                sep="\t",
                header=0,
                keep_default_na=False,
            )
            .reset_index(names="qname")
            .merge(
                right=df_sam,
                how="left",
                on=["qname"],
                validate="one_to_one",
            )[
                [
                    "R1_barcode",
                    "R1_primer",
                    "G",
                    "R1_sgRNA",
                    "R1_scaffold_prefix",
                    "R1_tail",
                    "R1_primer_score",
                    "R1_scaffold_prefix_score",
                    "R2_barcode",
                    "R2_primer",
                    "barcode_head",
                    "barcode",
                    "CTG_target_prefix",
                    "R2_sgRNA",
                    "pam",
                    "target_suffix",
                    "C",
                    "R2_scaffold_prefix",
                    "R2_tail",
                    "R2_primer_score",
                    "R2_scaffold_prefix_score",
                    "R2_sgRNA_score",
                    "barcode_score",
                    "barcode_id",
                    "count",
                ]
            ]
        )

        df_parse.to_csv(
            root_dir / "parse" / "bar" / parse_file,
            sep="\t",
            index=False,
        )


def build_sgRNA(root_dir: os.PathLike, plasmid_file: os.PathLike | None):
    root_dir = pathlib.Path(os.fspath(root_dir))
    os.makedirs(root_dir / "sgRNA" / "index", exist_ok=True)
    if plasmid_file is None:
        with resources.as_file(
            resources.files().parent
            / "plasmids"
            / "plasmids/final_hgsgrna_libb_all_0811_NAA_scaffold_nbt.csv"
        ) as pf:
            df_plasmid = pd.read_csv(pf, header=0)
    else:
        df_plasmid = pd.read_csv(plasmid_file, header=0)

    with open(root_dir / "sgRNA" / "index" / "sgRNA.fa", "w") as fd:
        for i, sgRNA in enumerate(df_plasmid["sgRNA"]):
            fd.write(f">b{i}\n{sgRNA}\n")

        subprocess.run(
            args=[
                "bowtie2-build",
                (root_dir / "sgRNA" / "index" / "sgRNA.fa").as_posix(),
                (root_dir / "sgRNA" / "index" / "sgRNA").as_posix(),
            ]
        )


def prepare_R1_sgRNA_and_R2_sgRNA(root_dir: os.PathLike):
    root_dir = pathlib.Path(os.fspath(root_dir))
    for column in ["R1_sgRNA", "R2_sgRNA"]:
        os.makedirs(root_dir / "sgRNA" / "read" / column, exist_ok=True)
        for parse_file in os.listdir(root_dir / "parse" / "bar"):
            df = pd.read_csv(
                root_dir / "parse" / "bar" / parse_file,
                sep="\t",
                header=0,
                usecols=[column],
                keep_default_na=False,
            )
            df[column] = df[column].where(df[column] != "", "N")
            df.to_csv(
                root_dir
                / "sgRNA"
                / "read"
                / column
                / f"{pathlib.Path(parse_file).stem}.csv",
                header=False,
                index=False,
            )


def map_sgRNA(root_dir: os.PathLike):
    root_dir = pathlib.Path(os.fspath(root_dir))
    for column in ["R1_sgRNA", "R2_sgRNA"]:
        os.makedirs(root_dir / "sgRNA" / "align" / column, exist_ok=True)
        for csv_file in os.listdir(root_dir / "sgRNA" / "read" / column):
            subprocess.run(
                args=[
                    "bowtie2",
                    "--quiet",
                    "-p",
                    "24",
                    "--mm",
                    "--norc",
                    "--local",
                    "-L",
                    "7",
                    "--ma",
                    "1",
                    "--mp",
                    "2,2",
                    "--rdg",
                    "3,1",
                    "--rfg",
                    "3,1",
                    "--score-min",
                    "C,1",
                    "-r",
                    "-x",
                    (root_dir / "sgRNA" / "index" / "sgRNA").as_posix(),
                    "-U",
                    (root_dir / "sgRNA" / "read" / column / csv_file).as_posix(),
                    "-S",
                    (
                        root_dir
                        / "sgRNA"
                        / "align"
                        / column
                        / f"{pathlib.Path(csv_file).stem}.sam"
                    ).as_posix(),
                ],
            )


def parse_sgRNA(root_dir: os.PathLike):
    root_dir = pathlib.Path(os.fspath(root_dir))
    os.makedirs(root_dir / "parse" / "sgRNA", exist_ok=True)
    for parse_file in os.listdir(root_dir / "parse" / "bar"):
        sam_file = f"{pathlib.Path(parse_file).stem}.sam"

        R1_sam = pysam.AlignmentFile(
            root_dir / "sgRNA" / "align" / "R1_sgRNA" / sam_file
        )
        R1_qnames = []
        R1_sgRNA_ids = []
        R1_sgRNA_bowtie2_scores = []
        for align in R1_sam.fetch():
            R1_qnames.append(int(align.qname))
            if (align.flag // 4) % 2 == 1:
                R1_sgRNA_ids.append(-1)
                R1_sgRNA_bowtie2_scores.append(0)
            else:
                R1_sgRNA_ids.append(align.rname)
                R1_sgRNA_bowtie2_scores.append(align.get_tag("AS"))
        df_R1_sam = pd.DataFrame(
            {
                "qname": R1_qnames,
                "R1_sgRNA_id": R1_sgRNA_ids,
                "R1_sgRNA_bowtie2_score": R1_sgRNA_bowtie2_scores,
            }
        )

        R2_sam = pysam.AlignmentFile(
            root_dir / "sgRNA" / "align" / "R2_sgRNA" / sam_file
        )
        R2_qnames = []
        R2_sgRNA_ids = []
        R2_sgRNA_bowtie2_scores = []
        for align in R2_sam.fetch():
            R2_qnames.append(int(align.qname))
            if (align.flag // 4) % 2 == 1:
                R2_sgRNA_ids.append(-1)
                R2_sgRNA_bowtie2_scores.append(0)
            else:
                R2_sgRNA_ids.append(align.rname)
                R2_sgRNA_bowtie2_scores.append(align.get_tag("AS"))
        df_R2_sam = pd.DataFrame(
            {
                "qname": R2_qnames,
                "R2_sgRNA_id": R2_sgRNA_ids,
                "R2_sgRNA_bowtie2_score": R2_sgRNA_bowtie2_scores,
            }
        )

        df_parse = (
            pd.read_csv(
                root_dir / "parse" / "bar" / parse_file,
                sep="\t",
                header=0,
                keep_default_na=False,
            )
            .reset_index(names="qname")
            .merge(
                right=df_R1_sam,
                how="left",
                on=["qname"],
                validate="one_to_one",
            )
            .merge(
                right=df_R2_sam,
                how="left",
                on=["qname"],
                validate="one_to_one",
            )[
                [
                    "R1_barcode",
                    "R1_primer",
                    "G",
                    "R1_sgRNA",
                    "R1_scaffold_prefix",
                    "R1_tail",
                    "R1_primer_score",
                    "R1_scaffold_prefix_score",
                    "R2_barcode",
                    "R2_primer",
                    "barcode_head",
                    "barcode",
                    "CTG_target_prefix",
                    "R2_sgRNA",
                    "pam",
                    "target_suffix",
                    "C",
                    "R2_scaffold_prefix",
                    "R2_tail",
                    "R2_primer_score",
                    "R2_scaffold_prefix_score",
                    "R2_sgRNA_score",
                    "barcode_score",
                    "R1_sgRNA_bowtie2_score",
                    "R2_sgRNA_bowtie2_score",
                    "barcode_id",
                    "R1_sgRNA_id",
                    "R2_sgRNA_id",
                    "count",
                ]
            ]
        )

        df_parse.to_csv(
            root_dir / "parse" / "sgRNA" / parse_file,
            sep="\t",
            index=False,
        )



import os
import pathlib

import pandas as pd

from . import utils


def duplicate_treat(root_dir: os.PathLike):
    root_dir = pathlib.Path(os.fspath(root_dir))
    os.makedirs(root_dir / "analyze" / "treat" / "dup", exist_ok=True)
    df_treat = pd.read_feather(
        root_dir / "analyze" / "treat" / "filter" / "mutant" / "treat.feather"
    )
    df_treat = df_treat.assign(
        chip=lambda df: df["stem"].map(utils.infer_chip),
        time=lambda df: df["stem"].map(utils.infer_time),
        wt="wt1 wt2",
    )
    df_treat.loc[(df_treat["chip"] == "g2n") & (df_treat["time"] == 2), "wt"] = (
        "wt1 wt2 wt11 wt21"
    )
    (
        df_treat.assign(wt=lambda df: df["wt"].str.split())
        .explode("wt")
        .reset_index(drop=True)
        .assign(stem=lambda df: df["stem"] + "_" + df["wt"])
        .to_feather(root_dir / "analyze" / "treat" / "dup" / "treat.feather")
    )


def duplicate_control(root_dir: os.PathLike):
    root_dir = pathlib.Path(os.fspath(root_dir))
    os.makedirs(root_dir / "analyze" / "control" / "dup", exist_ok=True)
    df_control = pd.read_feather(
        root_dir / "analyze" / "control" / "full" / "control.feather"
    )

    df_control = df_control.assign(
        chip=lambda df: df["stem"].map(utils.infer_chip),
        time=lambda df: df["stem"].map(utils.infer_time).astype(str),
        wt=lambda df: df["stem"].map(utils.infer_wt),
    )

    df_control.loc[
        (df_control["wt"] == "wt1")
        & (df_control["chip"] == "a2")
        & (df_control["time"] == "1"),
        "time",
    ] = "1 3"
    df_control.loc[
        (df_control["wt"] == "wt1")
        & (df_control["chip"] == "a2")
        & (df_control["time"] == "2"),
        "time",
    ] = "2 4"
    df_control.loc[
        (df_control["wt"] == "wt1")
        & (df_control["chip"] == "g1n")
        & (df_control["time"] == "2"),
        "time",
    ] = "2 1"
    df_control.loc[
        (df_control["wt"] == "wt1")
        & (df_control["chip"] == "g2n")
        & (df_control["time"] == "4"),
        "time",
    ] = "4 3"
    df_control.loc[
        (df_control["wt"] == "wt1")
        & (df_control["chip"] == "g3n")
        & (df_control["time"] == "1"),
        "time",
    ] = "1 2 3 4"
    df_control.loc[
        (df_control["wt"] == "wt2")
        & (df_control["chip"] == "a2")
        & (df_control["time"] == "3"),
        "time",
    ] = "3 4"
    df_control.loc[
        (df_control["wt"] == "wt2")
        & (df_control["chip"] == "a3")
        & (df_control["time"] == "3"),
        "time",
    ] = "3 4"
    df_control.loc[
        (df_control["wt"] == "wt2")
        & (df_control["chip"] == "g3n")
        & (df_control["time"] == "1"),
        "time",
    ] = "1 2 3 4"

    (
        df_control.assign(time=lambda df: df["time"].str.split())
        .explode("time")
        .reset_index(drop=True)
        .assign(stem=lambda df: df["stem"] + "_" + df["time"])
        .astype({"time": int})
        .to_feather(root_dir / "analyze" / "control" / "dup" / "control.feather")
    )


def merge(root_dir: os.PathLike):
    """
    count_wt_ctl and count_tot_ctl must be calculated before left merge because information about control will loss after left merge.
    """
    root_dir = pathlib.Path(os.fspath(root_dir))
    os.makedirs(root_dir / "analyze" / "treat" / "merge", exist_ok=True)
    df_treat = pd.read_feather(root_dir / "analyze" / "treat" / "dup" / "treat.feather")
    df_control = (
        pd.read_feather(root_dir / "analyze" / "control" / "dup" / "control.feather")
        .assign(
            count_wt_ctl=lambda df: utils.count_wt(df),
            count_tot_ctl=lambda df: df.groupby(["stem", "ref_id"])["count"].transform(
                "sum"
            ),
        )
        .rename(columns={"count": "count_ctl"})
    )

    on = [
        "chip",
        "time",
        "wt",
        "ref_id",
        "ref_end1",
        "ref_start2",
        "random_insertion",
    ]
    df_treat = df_treat.merge(
        right=df_control[on + ["count_ctl"]],
        how="left",
        on=on,
        validate="many_to_one",
    ).assign(count_ctl=lambda df: df["count_ctl"].fillna(0))

    on = [
        "chip",
        "time",
        "wt",
        "ref_id",
    ]
    df_treat = df_treat.merge(
        right=df_control.groupby(on)
        .agg(
            count_wt_ctl=pd.NamedAgg(column="count_wt_ctl", aggfunc="first"),
            count_tot_ctl=pd.NamedAgg(column="count_tot_ctl", aggfunc="first"),
        )
        .reset_index(),
        how="left",
        on=on,
        validate="many_to_one",
    ).assign(
        count_wt_ctl=lambda df: df["count_wt_ctl"].fillna(0),
        count_tot_ctl=lambda df: df["count_tot_ctl"].fillna(0),
    )

    df_treat.to_feather(root_dir / "analyze" / "treat" / "merge" / "treat.feather")
